const builtin = @import("builtin");

const io = @import("./io.zig");

const std = @import("std");

pub const AccessError = error {
	FileNotFound,
	FileInaccessible,
};

pub const ReadAllError = AccessError || error {
	ReadIncomplete,
};

pub const ReadAllOptions = struct {
	offset: u64 = 0,
	limit: u64 = std.math.maxInt(u64),
};

pub const Stat = struct {
	size: u64,
};

pub const Storage = struct {
	userdata: *anyopaque,
	vtable: *const VTable,

	pub const VTable = struct {
		stat: *const fn (*anyopaque, []const u8) AccessError!Stat,
		read: *const fn (*anyopaque, []const u8, usize, []io.Byte) AccessError!usize,
	};

	pub fn read(self: Storage, path: []const u8, output: []io.Byte, offset: u64) AccessError!usize {
		return self.vtable.read(self.userdata, path, offset, output);
	}

	pub fn read_all(self: Storage, path: []const u8, output: []io.Byte, options: ReadAllOptions) ReadAllError![]const io.Byte {
		const bytes_read = try self.vtable.read(self.userdata, path, options.offset, output);

		if (try self.vtable.read(self.userdata, path, options.offset, output) != output.len) {
			return error.ReadIncomplete;
		}

		return output[0 .. bytes_read];
	}

	pub fn read_alloc(self: Storage, path: []const u8, allocator: std.mem.Allocator, options: ReadAllOptions) (std.mem.Allocator.Error || ReadAllError)![]io.Byte {
		const buffer = try allocator.alloc(io.Byte, @min((try self.stat(path)).size, options.limit));

		errdefer allocator.free(buffer);

		if (try self.vtable.read(self.userdata, path, options.offset, buffer) != buffer.len) {
			return error.ReadIncomplete;
		}

		return buffer;
	}

	pub fn read_foreign(self: Storage, path: []const u8, offset: u64, comptime Type: type) AccessError!?Type {
		const decoded = (try self.read_native(path, offset, Type)) orelse {
			return null;
		};

		return switch (@typeInfo(Type)) {
			.Struct => std.mem.byteSwapAllFields(Type, &decoded),
			else => @byteSwap(decoded),
		};
	}

	pub fn read_native(self: Storage, path: []const u8, offset: u64, comptime Type: type) AccessError!?Type {
		var buffer = @as([@sizeOf(Type)]io.Byte, undefined);

		if (try self.vtable.read(self.userdata, path, offset, &buffer) != buffer.len) {
			return null;
		}

		return @as(*align(1) const Type, @ptrCast(&buffer)).*;
	}

	pub fn stat(self: Storage, path: []const u8) AccessError!Stat {
		return self.vtable.stat(self.userdata, path);
	}

	pub const read_little = switch (native_endian) {
		.little => read_native,
		.big => read_foreign,
	};

	pub const read_big = switch (native_endian) {
		.little => read_foreign,
		.big => read_native,
	};
};

pub const bundle = init: {
	const Bundle = struct {
		fn full_path(path: []const u8) AccessError![4095:0]u8 {
			var buffer = [_:0]u8{0} ** 4095;

			_ = std.fs.cwd().realpath(path, &buffer) catch {
				return error.FileInaccessible;
			};

			return buffer;
		}

		fn read(_: *anyopaque, path: []const u8, offset: usize, output: []io.Byte) AccessError!usize {
			var file = std.fs.openFileAbsoluteZ(&(try full_path(path)), .{.mode = .read_only}) catch |open_error| {
				return switch (open_error) {
					error.FileNotFound => error.FileNotFound,
					else => error.FileInaccessible,
				};
			};

			defer file.close();

			if (offset != 0) {
				file.seekTo(offset) catch {
					return error.FileInaccessible;
				};
			}

			return file.read(output) catch error.FileInaccessible;
		}

		fn stat(_: *anyopaque, path: []const u8) AccessError!Stat {
			const file_stat = get: {
				var file = std.fs.openFileAbsoluteZ(&(try full_path(path)), .{.mode = .read_only}) catch |open_error| {
					return switch (open_error) {
						error.FileNotFound => error.FileNotFound,
						else => error.FileInaccessible,
					};
				};

				defer file.close();

				break: get file.stat() catch {
					return error.FileInaccessible;
				};
			};

			return .{
				.size = file_stat.size,
			};
		}
	};

	break: init Storage{
		.userdata = undefined,

		.vtable = &.{
			.stat = Bundle.stat,
			.read = Bundle.read,
		},
	};
};

const native_endian = builtin.cpu.arch.endian();
