const builtin = @import("builtin");

const flow = @import("./flow.zig");

const coral = @import("coral");

const states = @import("./states.zig");

const std = @import("std");

const system = @import("./system.zig");

thread_pool: ?*std.Thread.Pool = null,
thread_restricted_resources: [std.enums.values(states.ThreadRestriction).len]states.Table,
event_systems: coral.stack.Sequential(system.Schedule),

pub const Event = enum (usize) { _ };

const Self = @This();

pub fn create_event(self: *Self, label: []const u8) std.mem.Allocator.Error!Event {
	var systems = try system.Schedule.init(label);

	errdefer systems.deinit(self);

	const index = self.event_systems.len();

	try self.event_systems.push_grow(systems);

	return @enumFromInt(index);
}

pub fn deinit(self: *Self) void {
	for (self.event_systems.values) |*schedule| {
		schedule.deinit(self);
	}

	for (&self.thread_restricted_resources) |*resources| {
		resources.deinit();
	}

	if (self.thread_pool) |thread_pool| {
		thread_pool.deinit();
		coral.heap.allocator.destroy(thread_pool);
	}

	self.event_systems.deinit();

	self.* = undefined;
}

pub fn get_params(self: Self, comptime Value: type) flow.Params(Value) {
	const params = self.get_state(flow.Params(Value)) orelse {
		return .{};
	};

	return params.*;
}

pub fn get_state(self: Self, comptime Value: type) ?*Value {
	return @ptrCast(@alignCast(self.thread_restricted_resources[@intFromEnum(states.thread_restriction(Value))].get(Value)));
}

pub fn set_get_state(self: *Self, value: anytype) std.mem.Allocator.Error!*@TypeOf(value) {
	return self.thread_restricted_resources[@intFromEnum(states.thread_restriction(@TypeOf(value)))].set_get(value);
}

pub fn init(thread_count: u32) std.Thread.SpawnError!Self {
	var world = Self{
		.thread_restricted_resources = .{states.Table.init(), states.Table.init()},
		.event_systems = .{.allocator = coral.heap.allocator},
	};

	if (thread_count != 0 and !builtin.single_threaded) {
		const thread_pool = try coral.heap.allocator.create(std.Thread.Pool);

		try thread_pool.init(.{
			.allocator = coral.heap.allocator,
			.n_jobs = thread_count,
		});

		world.thread_pool = thread_pool;
	}

	return world;
}

pub fn on_event(self: *Self, event: Event, action: *const system.Info, order: system.Order) std.mem.Allocator.Error!void {
	try self.event_systems.values[@intFromEnum(event)].then(self, action, order);
}

pub fn run_event(self: *Self, event: Event) anyerror!void {
	try self.event_systems.values[@intFromEnum(event)].run(self);
}

pub fn set_state(self: *Self, value: anytype) std.mem.Allocator.Error!void {
	try self.thread_restricted_resources[@intFromEnum(states.thread_restriction(@TypeOf(value)))].set(value);
}
