const coral = @import("coral");

const gfx = @import("../gfx.zig");

const sokol = @import("sokol");

const spirv = @import("./spirv.zig");

const std = @import("std");

effects: EffectPool,
textures: TexturePool,

pub const Effect = struct {
	shader: sokol.gfx.Shader,
	pipeline: sokol.gfx.Pipeline,
	properties: []coral.io.Byte,

	pub fn deinit(self: *Effect) void {
		coral.heap.allocator.free(self.properties);
		sokol.gfx.destroyPipeline(self.pipeline);
		sokol.gfx.destroyShader(self.shader);

		self.* = undefined;
	}

	pub fn init(desc: gfx.Effect.Desc) spirv.Error!Effect {
		var spirv_arena = std.heap.ArenaAllocator.init(coral.heap.allocator);

		defer {
			spirv_arena.deinit();
		}

		const spirv_program = try spirv.analyze(&spirv_arena, .{
			.target = try switch (sokol.gfx.queryBackend()) {
				.GLCORE => spirv.Target.glsl,
				else => error.InvalidSPIRV,
			},

			.vertex_source = .{
				.ops = &spirv.to_ops(@embedFile("./shaders/2d_default.vert.spv")),
			},

			.fragment_source = .{
				.ops = desc.fragment_spirv_ops,
			},
		});

		const shader = sokol.gfx.makeShader(shader_desc: {
			const shader_desc = sokol.gfx.ShaderDesc{
				.vs = stage_desc(spirv_program.vertex_stage),
				.fs = stage_desc(spirv_program.fragment_stage),
			};

			// TODO: Vertex attributes, for some reason they aren't needed?

			break: shader_desc shader_desc;
		});

		// TODO: Review blending rules.
		const pipeline = sokol.gfx.makePipeline(pipeline_desc: {
			var pipeline_desc = sokol.gfx.PipelineDesc{
				.label = "Effect pipeline",
				.layout = vertex_layout_state,
				.shader = shader,
				.index_type = .UINT16,
				.blend_color = .{.r = 1.0, .g = 1.0, .b = 1.0, .a = 1.0},
			};

			pipeline_desc.colors[0] = .{
				.write_mask = .RGBA,

				.blend = .{
					.enabled = true,
					.src_factor_rgb = .SRC_ALPHA,
					.dst_factor_rgb = .ONE_MINUS_SRC_ALPHA,
				},
			};

			break: pipeline_desc pipeline_desc;
		});

		const properties = try coral.heap.allocator.alloc(
			coral.io.Byte,
			if (spirv_program.fragment_stage.has_uniform_blocks[0]) |uniform_block| uniform_block.size() else 0,
		);

		errdefer {
			coral.heap.allocator.free(properties);
			sokol.gfx.destroyPipeline(pipeline);
			sokol.gfx.destroyShader(shader);
		}

		return .{
			.shader = shader,
			.pipeline = pipeline,
			.properties = properties,
		};
	}

	fn stage_desc(spirv_stage: spirv.Stage) sokol.gfx.ShaderStageDesc {
		var stage = sokol.gfx.ShaderStageDesc{
			.entry = spirv_stage.entry_point,
			.source = spirv_stage.source,
		};

		for (0 .. spirv.Stage.max_uniform_blocks) |slot| {
			const uniform_block = &(spirv_stage.has_uniform_blocks[slot] orelse {
				continue;
			});

			const stage_uniform_block = &stage.uniform_blocks[slot];

			stage_uniform_block.layout = switch (uniform_block.layout) {
				.std140 => .STD140,
			};

			stage_uniform_block.size = uniform_block.size();

			for (stage_uniform_block.uniforms[0 .. uniform_block.uniforms.len], uniform_block.uniforms) |*stage_uniform, uniform| {
				stage_uniform.* = .{
					.type = switch (uniform.type) {
						.float => .FLOAT,
						.float2 => .FLOAT2,
						.float3 => .FLOAT3,
						.float4 => .FLOAT4,
						.integer => .INT,
						.integer2 => .INT2,
						.integer3 => .INT3,
						.integer4 => .INT4,
						.matrix4 => .MAT4,
					},

					.name = uniform.name,
					.array_count = uniform.len,
				};
			}
		}

		for (0 .. spirv.Stage.max_images) |slot| {
			const image = &(spirv_stage.has_images[slot] orelse {
				continue;
			});

			stage.images[slot] = .{
				.multisampled = image.is_multisampled,
				.image_type = ._2D,
				.sample_type = .FLOAT,
				.used = true,
			};
		}

		for (0 .. spirv.Stage.max_samplers) |slot| {
			const sampler = &(spirv_stage.has_samplers[slot] orelse {
				continue;
			});

			stage.samplers[slot] = .{
				.sampler_type = switch (sampler.*) {
					.filtering => .FILTERING,
					.non_filtering => .NONFILTERING,
					.comparison => .COMPARISON,
				},

				.used = true,
			};
		}

		for (0 .. spirv.Stage.max_image_sampler_pairs) |slot| {
			const image_sampler_pair = &(spirv_stage.has_image_sampler_pairs[slot] orelse {
				continue;
			});

			stage.image_sampler_pairs[slot] = .{
				.glsl_name = image_sampler_pair.name,
				.image_slot = @intCast(image_sampler_pair.image_slot),
				.sampler_slot = @intCast(image_sampler_pair.sampler_slot),
				.used = true,
			};
		}

		return stage;
	}

	const vertex_layout_state = sokol.gfx.VertexLayoutState{
		.attrs = get: {
			var attrs = [_]sokol.gfx.VertexAttrState{.{}} ** 16;

			attrs[0] = .{
				.format = .FLOAT2,
				.buffer_index = 0,
			};

			attrs[1] = .{
				.format = .FLOAT2,
				.buffer_index = 0,
			};

			attrs[2] = .{
				.format = .FLOAT2,
				.buffer_index = 1,
			};

			attrs[3] = .{
				.format = .FLOAT2,
				.buffer_index = 1,
			};

			attrs[4] = .{
				.format = .FLOAT2,
				.buffer_index = 1,
			};

			attrs[5] = .{
				.format = .UBYTE4N,
				.buffer_index = 1,
			};

			attrs[6] = .{
				.format = .FLOAT,
				.buffer_index = 1,
			};

			attrs[7] = .{
				.format = .FLOAT4,
				.buffer_index = 1,
			};

			break: get attrs;
		},

		.buffers = get: {
			var vertex_buffer_layouts = [_]sokol.gfx.VertexBufferLayoutState{.{}} ** 8;

			vertex_buffer_layouts[1].step_func = .PER_INSTANCE;

			break: get vertex_buffer_layouts;
		},
	};
};

const EffectPool = coral.Pool(Effect);

const Self = @This();

pub const Texture = struct {
	width: u16,
	height: u16,
	access: Access,

	pub const Access = union (enum) {
		empty,
		render: RenderAccess,
		static: StaticAccess,
	};

	pub const RenderAccess = struct {
		color_image: sokol.gfx.Image,
		depth_image: sokol.gfx.Image,
		attachments: sokol.gfx.Attachments,
	};

	pub const StaticAccess = struct {
		image: sokol.gfx.Image,
	};

	pub fn deinit(self: *Texture) void {
		switch (self.access) {
			.render => |render| {
				sokol.gfx.destroyImage(render.color_image);
				sokol.gfx.destroyImage(render.depth_image);
				sokol.gfx.destroyAttachments(render.attachments);
			},

			.static => |static| {
				sokol.gfx.destroyImage(static.image);
			},

			.empty => {},
		}

		self.* = undefined;
	}

	pub fn init(desc: gfx.Texture.Desc) std.mem.Allocator.Error!Texture {
		const pixel_format = switch (desc.format) {
			.rgba8 => sokol.gfx.PixelFormat.RGBA8,
			.bgra8 => sokol.gfx.PixelFormat.BGRA8,
		};

		switch (desc.access) {
			.render => |render| {
				if (render.width == 0 or render.height == 0) {
					return .{
						.width = render.width,
						.height = render.height,
						.access = .empty,
					};
				}

				const color_image = sokol.gfx.makeImage(.{
					.pixel_format = pixel_format,
					.width = render.width,
					.height = render.height,
					.render_target = true,
				});

				const depth_image = sokol.gfx.makeImage(.{
					.width = render.width,
					.height = render.height,
					.render_target = true,
					.pixel_format = .DEPTH_STENCIL,
				});

				const attachments = sokol.gfx.makeAttachments(attachments_desc: {
					var attachments_desc = sokol.gfx.AttachmentsDesc{
						.depth_stencil = .{
							.image = depth_image,
						},
					};

					attachments_desc.colors[0] = .{
						.image = color_image,
					};

					break: attachments_desc attachments_desc;
				});

				return .{
					.width = render.width,
					.height = render.height,

					.access = .{
						.render = .{
							.attachments = attachments,
							.color_image = color_image,
							.depth_image = depth_image,
						},
					},
				};
			},

			.static => |static| {
				const height = std.math.cast(u16, static.data.len / (static.width * desc.format.byte_size())) orelse {
					return error.OutOfMemory;
				};

				if (static.width == 0 or height == 0) {
					return .{
						.width = static.width,
						.height = height,
						.access = .empty,
					};
				}

				const image = sokol.gfx.makeImage(image_desc: {
					var image_desc = sokol.gfx.ImageDesc{
						.height = height,
						.pixel_format = pixel_format,
						.width = static.width,
					};

					image_desc.data.subimage[0][0] = sokol.gfx.asRange(static.data);

					break: image_desc image_desc;
				});

				errdefer {
					sokol.gfx.destroyImage(image);
				}

				return .{
					.width = static.width,
					.height = height,

					.access = .{
						.static = .{
							.image = image,
						},
					},
				};
			},
		}
	}
};

const TexturePool = coral.Pool(Texture);

pub fn create_effect(self: *Self, desc: gfx.Effect.Desc) !gfx.Effect {
	var effect = try Effect.init(desc);

	errdefer effect.deinit();

	return @enumFromInt(try self.effects.insert(effect));
}

pub fn create_texture(self: *Self, desc: gfx.Texture.Desc) !gfx.Texture {
	var texture = try Texture.init(desc);

	errdefer texture.deinit();

	return @enumFromInt(try self.textures.insert(texture));
}

pub fn deinit(self: *Self) void {
	var textures = self.textures.values();

	while (textures.next()) |texture| {
		texture.deinit();
	}

	self.textures.deinit();

	var effects = self.effects.values();

	while (effects.next()) |effect| {
		effect.deinit();
	}

	self.effects.deinit();

	self.* = undefined;
}

pub fn destroy_effect(self: *Self, handle: gfx.Effect) bool {
	switch (handle) {
		.default => {},

		else => {
			var effect = self.effects.remove(@intFromEnum(handle)) orelse {
				return false;
			};

			effect.deinit();
		},
	}

	return true;
}

pub fn destroy_texture(self: *Self, handle: gfx.Texture) bool {
	switch (handle) {
		.default => {},

		else => {
			var texture = self.textures.remove(@intFromEnum(handle)) orelse {
				return false;
			};

			texture.deinit();
		},
	}

	return true;
}

pub fn get_effect(self: *Self, handle: gfx.Effect) ?*Effect {
	return self.effects.get(@intFromEnum(handle));
}

pub fn get_texture(self: *Self, handle: gfx.Texture) ?*Texture {
	return self.textures.get(@intFromEnum(handle));
}

pub fn init() !Self {
	var pools = Self{
		.effects = EffectPool.init(coral.heap.allocator),
		.textures = TexturePool.init(coral.heap.allocator),
	};

	errdefer {
		pools.deinit();
	}

	const assert = struct {
		fn is_handle(expected: anytype, actual: @TypeOf(expected)) void {
			std.debug.assert(actual == expected);
		}
	};

	assert.is_handle(gfx.Effect.default, try pools.create_effect(.{
		.fragment_spirv_ops = &spirv.to_ops(@embedFile("./shaders/2d_default.frag.spv")),
	}));

	assert.is_handle(gfx.Texture.default, try pools.create_texture(.{
		.format = .rgba8,

		.access = .{
			.static = .{
				.data = std.mem.asBytes(&[_]u32{
					0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080,
					0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000,
					0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080,
					0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000,
					0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080,
					0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000,
					0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080,
					0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000, 0xFF800080, 0xFF000000,
				}),

				.width = 8,
			},
		},
	}));

	assert.is_handle(gfx.Texture.backbuffer, try pools.create_texture(.{
		.format = .rgba8,

		.access = .{
			.render = .{
				.width = 0,
				.height = 0,
			},
		}
	}));

	return pools;
}
