const coral = @import("coral");

const ext = @cImport({
	@cInclude("spirv-cross/spirv_cross_c.h");
});

const std = @import("std");

pub const Error = std.mem.Allocator.Error || error {
	InvalidSPIRV,
	UnsupportedSPIRV,
};

pub const Options = struct {
	target: Target,
	vertex_source: Source,
	fragment_source: Source,
};

pub const Program = struct {
	vertex_stage: Stage,
	fragment_stage: Stage,
};

pub const Source = struct {
	entry_point: []const u8 = "main",
	ops: []const u32,
};

pub const Stage = struct {
	entry_point: [:0]const u8,
	source: [:0]const u8,
	has_image_sampler_pairs: [max_image_sampler_pairs]?ImageSamplerPair = [_]?ImageSamplerPair{null} ** max_image_sampler_pairs,
	has_images: [max_images]?Image = [_]?Image{null} ** max_images,
    has_samplers: [max_samplers]?Sampler = [_]?Sampler{null} ** max_samplers,
	has_uniform_blocks: [max_uniform_blocks]?UniformBlock = [_]?UniformBlock{null} ** max_uniform_blocks,

	pub const Image = struct {
		is_multisampled: bool,
	};

	pub const ImageSamplerPair = struct {
		image_slot: usize,
		sampler_slot: usize,
		name: [:0]const u8,
	};

	pub const Layout = enum {
		std140,
	};

	pub const Sampler = enum {
		filtering,
		non_filtering,
		comparison,
	};

	pub const Uniform = struct {
		name: [:0]const u8,
		type: Type,
		len: u16,

		pub const Type = enum {
			float,
			float2,
			float3,
			float4,
			integer,
			integer2,
			integer3,
			integer4,
			matrix4,
		};
	};

	pub const UniformBlock = struct {
		layout: Layout,
		uniforms: []const Uniform,

		pub const max_uniforms = 16;

		pub fn size(self: UniformBlock) usize {
			var accumulated_size: usize = 0;

			for (self.uniforms) |uniform| {
				accumulated_size += @max(1, uniform.len) * @as(usize, switch (uniform.type) {
					.float => 4,
					.float2 => 8,
					.float3 => 12,
					.float4 => 16,
					.integer => 4,
					.integer2 => 8,
					.integer3 => 12,
					.integer4 => 16,
					.matrix4 => 64,
				});
			}

			const alignment: usize = switch (self.layout) {
				.std140 => 16,
			};

			return (accumulated_size + (alignment - 1)) & ~(alignment - 1);
		}
	};

	pub const max_images = 12;

	pub const max_image_sampler_pairs = 12;

	pub const max_samplers = 8;

	pub const max_storage_buffers = 8;

	pub const max_uniform_blocks = 8;

	fn reflect_images_samplers(self: *Stage, arena: *std.heap.ArenaAllocator, compiler: ext.spvc_compiler, resources: ext.spvc_resources) Error!void {
		const arena_allocator = arena.allocator();
		var sampled_images: []const ext.spvc_reflected_resource = &.{};

		try to_error(ext.spvc_resources_get_resource_list_for_type(
			resources,
			ext.SPVC_RESOURCE_TYPE_SAMPLED_IMAGE,
			@ptrCast(&sampled_images.ptr),
			&sampled_images.len,
		));

		if (sampled_images.len > max_image_sampler_pairs) {
			return error.UnsupportedSPIRV;
		}

		for (0 .. sampled_images.len, sampled_images) |i, sampled_image| {
			const sampled_image_type = ext.spvc_compiler_get_type_handle(compiler, sampled_image.type_id);

			if (ext.spvc_type_get_basetype(sampled_image_type) != ext.SPVC_BASETYPE_SAMPLED_IMAGE) {
				return error.InvalidSPIRV;
			}

			try switch (ext.spvc_type_get_image_dimension(sampled_image_type)) {
				ext.SpvDim2D => {},
				else => error.InvalidSPIRV,
			};

			try switch (ext.spvc_type_get_basetype(ext.spvc_compiler_get_type_handle(compiler, ext.spvc_type_get_image_sampled_type(sampled_image_type)))) {
				ext.SPVC_BASETYPE_FP32 => {},
				else => error.InvalidSPIRV,
			};

			self.has_images[i] = .{
				.is_multisampled = ext.spvc_type_get_image_multisampled(sampled_image_type) != 0,
			};

			self.has_samplers[i] = .filtering;

			self.has_image_sampler_pairs[i] = .{
				.name = try arena_allocator.dupeZ(u8, std.mem.span(ext.spvc_compiler_get_name(compiler, sampled_image.id))),
				.image_slot = @intCast(i),
				.sampler_slot = @intCast(i),
			};
		}
	}

	fn reflect_uniform_blocks(self: *Stage, arena: *std.heap.ArenaAllocator, compiler: ext.spvc_compiler, resources: ext.spvc_resources) Error!void {
		const arena_allocator = arena.allocator();
		var reflected_resources: []const ext.spvc_reflected_resource = &.{};

		try to_error(ext.spvc_resources_get_resource_list_for_type(
			resources,
			ext.SPVC_RESOURCE_TYPE_UNIFORM_BUFFER,
			@ptrCast(&reflected_resources.ptr),
			&reflected_resources.len,
		));

		if (reflected_resources.len > max_uniform_blocks) {
			return error.InvalidSPIRV;
		}

		for (0 .. reflected_resources.len, reflected_resources) |i, reflected_resource| {
			const type_handle = ext.spvc_compiler_get_type_handle(compiler, reflected_resource.base_type_id);
			const is_std430 = ext.spvc_compiler_has_decoration(compiler, reflected_resource.id, ext.SpvDecorationBufferBlock) == ext.SPVC_TRUE;

			if (is_std430) {
				return error.UnsupportedSPIRV;
			}

			const uniform_count = ext.spvc_type_get_num_member_types(type_handle);

			if (uniform_count > UniformBlock.max_uniforms) {
				return error.UnsupportedSPIRV;
			}

			const uniforms = try arena_allocator.alloc(Uniform, uniform_count);

			for (uniforms, 0 .. uniform_count) |*uniform, j| {
				const member_index: c_uint = @intCast(j);
				const member_type_handle = ext.spvc_compiler_get_type_handle(compiler, ext.spvc_type_get_member_type(type_handle, member_index));

				if (ext.spvc_type_get_num_array_dimensions(member_type_handle) > 1) {
					return error.UnsupportedSPIRV;
				}

				uniform.* = .{
					.name = try coral.utf8.alloc_formatted(arena_allocator, "_{id}.{member_name}", .{
						.id = reflected_resource.id,
						.member_name = std.mem.span(ext.spvc_compiler_get_member_name(compiler, reflected_resource.base_type_id, member_index)),
					}),

					.type = try switch (ext.spvc_type_get_basetype(member_type_handle)) {
						ext.SPVC_BASETYPE_FP32 => switch (ext.spvc_type_get_vector_size(member_type_handle)) {
							4 => switch (ext.spvc_type_get_columns(member_type_handle)) {
								4 => Uniform.Type.matrix4,
								1 => Uniform.Type.float4,
								else => error.UnsupportedSPIRV,
							},

							1 => Uniform.Type.float,
							2 => Uniform.Type.float2,
							3 => Uniform.Type.float3,
							else => error.UnsupportedSPIRV,
						},

						ext.SPVC_BASETYPE_INT32 => try switch (ext.spvc_type_get_vector_size(member_type_handle)) {
							1 => Uniform.Type.integer,
							2 => Uniform.Type.integer2,
							3 => Uniform.Type.integer3,
							4 => Uniform.Type.integer4,
							else => error.UnsupportedSPIRV,
						},

						else => error.UnsupportedSPIRV,
					},

					.len = std.math.cast(u16, ext.spvc_type_get_array_dimension(member_type_handle, 0)) orelse {
						return error.UnsupportedSPIRV;
					},
				};
			}

			self.has_uniform_blocks[i] = .{
				.uniforms = uniforms,
				.layout = .std140,
			};
		}
	}
};

pub const Target = enum {
	glsl,
};

pub fn analyze(arena: *std.heap.ArenaAllocator, options: Options) Error!Program {
	var context: ext.spvc_context = null;

	if (ext.spvc_context_create(&context) != ext.SPVC_SUCCESS) {
		return error.OutOfMemory;
	}

	defer {
		ext.spvc_context_destroy(context);
	}

	ext.spvc_context_set_error_callback(context, log_errors, null);

	const arena_allocator = arena.allocator();

	return .{
		.vertex_stage = vertex_stage: {
			const compiler = try parse(arena, context, options.target, options.vertex_source);

			var stage = Stage{
				.entry_point = try arena_allocator.dupeZ(u8, options.vertex_source.entry_point),
				.source = try compile(arena, compiler),
			};

			const resources = create: {
				var resources: ext.spvc_resources = null;

				try to_error(ext.spvc_compiler_create_shader_resources(compiler, &resources));

				break: create resources;
			};

			try stage.reflect_images_samplers(arena, compiler, resources);
			try stage.reflect_uniform_blocks(arena, compiler, resources);

			std.log.info("{s}", .{stage.source});

			break: vertex_stage stage;
		},

		.fragment_stage = fragment_stage: {
			const compiler = try parse(arena, context, options.target, options.fragment_source);

			var stage = Stage{
				.entry_point = try arena_allocator.dupeZ(u8, options.fragment_source.entry_point),
				.source = try compile(arena, compiler),
			};

			const resources = create: {
				var resources: ext.spvc_resources = null;

				try to_error(ext.spvc_compiler_create_shader_resources(compiler, &resources));

				break: create resources;
			};

			try stage.reflect_images_samplers(arena, compiler, resources);
			try stage.reflect_uniform_blocks(arena, compiler, resources);

			std.log.info("{s}", .{stage.source});

			break: fragment_stage stage;
		},
	};
}

fn compile(arena: *std.heap.ArenaAllocator, compiler: ext.spvc_compiler) Error![:0]const u8 {
	const arena_allocator = arena.allocator();
	var source: [*:0]const u8 = "";

	try to_error(ext.spvc_compiler_compile(compiler, @ptrCast(&source)));

	return arena_allocator.dupeZ(u8, std.mem.span(source));
}

fn log_errors(userdata: ?*anyopaque, error_message: [*c]const u8) callconv(.C) void {
	std.debug.assert(userdata == null);
	std.log.err("{s}", .{error_message});
}

fn parse(arena: *std.heap.ArenaAllocator, context: ext.spvc_context, target: Target, source: Source) Error!ext.spvc_compiler {
	var parsed_ir: ext.spvc_parsed_ir = null;

	try to_error(ext.spvc_context_parse_spirv(context, source.ops.ptr, source.ops.len, &parsed_ir));

	var compiler: ext.spvc_compiler = null;

	const spvc_target = switch (target) {
		.glsl => ext.SPVC_BACKEND_GLSL,
	};

	try to_error(ext.spvc_context_create_compiler(context, spvc_target, parsed_ir, ext.SPVC_CAPTURE_MODE_TAKE_OWNERSHIP, &compiler));

	try to_error(ext.spvc_compiler_build_combined_image_samplers(compiler));

	var combined_image_samplers: []const ext.spvc_combined_image_sampler = &.{};

	try to_error(ext.spvc_compiler_get_combined_image_samplers(compiler, @ptrCast(&combined_image_samplers.ptr), &combined_image_samplers.len));

	const arena_allocator = arena.allocator();
	var binding: u32 = 0;

	for (combined_image_samplers) |combined_image_sampler| {
		const name = try coral.utf8.alloc_formatted(arena_allocator, "{image_name}_{sampler_name}", .{
			.image_name = std.mem.span(ext.spvc_compiler_get_name(compiler, combined_image_sampler.image_id)),
			.sampler_name = std.mem.span(ext.spvc_compiler_get_name(compiler, combined_image_sampler.sampler_id)),
		});

		ext.spvc_compiler_set_name(compiler, combined_image_sampler.combined_id, name);
		ext.spvc_compiler_set_decoration(compiler, combined_image_sampler.combined_id, ext.SpvDecorationBinding, binding);

		binding += 1;
	}

	const option_values: []const struct {ext.spvc_compiler_option, c_uint} = switch (target) {
		.glsl => &.{
			.{ext.SPVC_COMPILER_OPTION_GLSL_VERSION, 430},
			.{ext.SPVC_COMPILER_OPTION_GLSL_ES, @intFromBool(false)},
			.{ext.SPVC_COMPILER_OPTION_GLSL_VULKAN_SEMANTICS, @intFromBool(false)},
			.{ext.SPVC_COMPILER_OPTION_GLSL_EMIT_UNIFORM_BUFFER_AS_PLAIN_UNIFORMS, @intFromBool(true)},
			.{ext.SPVC_COMPILER_OPTION_FLIP_VERTEX_Y, @intFromBool(true)},
		},
	};

	var compiler_options: ext.spvc_compiler_options = null;

	try to_error(ext.spvc_compiler_create_compiler_options(compiler, &compiler_options));

	for (option_values) |option_value| {
		const entry, const value = option_value;

		try to_error(ext.spvc_compiler_options_set_uint(compiler_options, entry, value));
	}

	try to_error(ext.spvc_compiler_install_compiler_options(compiler, compiler_options));

	return compiler;
}

fn to_error(result: ext.spvc_result) Error!void {
	return switch (result) {
		ext.SPVC_SUCCESS => {},
		ext.SPVC_ERROR_INVALID_SPIRV => error.InvalidSPIRV,
		ext.SPVC_ERROR_UNSUPPORTED_SPIRV => error.UnsupportedSPIRV,
		ext.SPVC_ERROR_OUT_OF_MEMORY => error.OutOfMemory,
		ext.SPVC_ERROR_INVALID_ARGUMENT, ext.SPVC_ERROR_INT_MAX => unreachable,
		else => unreachable,
	};
}

pub fn to_ops(raw: anytype) [raw.len / @alignOf(u32)]u32 {
	var ops: [raw.len / @alignOf(u32)]u32 = undefined;

	@memcpy(std.mem.sliceAsBytes(&ops), raw);

	return ops;
}
